//	DrawingPanel.cpp - The main drawing panel.
//	----------------------------------------------------------------------------
//	This file is part of Motion Composition Tool, a simple application for
//	creating simple abstract animations.
//	Copyright (c) 2007 Niall Moody.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//	----------------------------------------------------------------------------

#include "DrawingPanel.h"

#include <iostream>

using namespace std;

//------------------------------------------------------------------------------
DrawingPanel::DrawingPanel():
Component(),
length(180.0f),
currentPosition(0.0f),
draggedPoly(-1),
draggedPoint(0),
showPoints(true)
{
	int i, j;
	Polygon tempPoly;

	tempPoly.time = 0.0f;

	tempPoly.x[0] = 0.1f;
	tempPoly.y[0] = 0.9f;
	tempPoly.x[1] = 0.2f;
	tempPoly.y[1] = 0.9f;
	tempPoly.x[2] = 0.2f;
	tempPoly.y[2] = 0.8f;
	tempPoly.x[3] = 0.1f;
	tempPoly.y[3] = 0.8f;

	polygons[0].insert(polygons[0].end(), 1, tempPoly);
	colours[0] = Colours::lightblue;

	tempPoly.x[0] = 0.8f;
	tempPoly.y[0] = 0.9f;
	tempPoly.x[1] = 0.9f;
	tempPoly.y[1] = 0.9f;
	tempPoly.x[2] = 0.9f;
	tempPoly.y[2] = 0.8f;
	tempPoly.x[3] = 0.8f;
	tempPoly.y[3] = 0.8f;

	polygons[1].insert(polygons[1].end(), 1, tempPoly);
	colours[1] = Colours::lightblue;

	tempPoly.x[0] = 0.8f;
	tempPoly.y[0] = 0.2f;
	tempPoly.x[1] = 0.9f;
	tempPoly.y[1] = 0.2f;
	tempPoly.x[2] = 0.9f;
	tempPoly.y[2] = 0.1f;
	tempPoly.x[3] = 0.8f;
	tempPoly.y[3] = 0.1f;

	polygons[2].insert(polygons[2].end(), 1, tempPoly);
	colours[2] = Colours::lightblue;

	tempPoly.x[0] = 0.1f;
	tempPoly.y[0] = 0.2f;
	tempPoly.x[1] = 0.2f;
	tempPoly.y[1] = 0.2f;
	tempPoly.x[2] = 0.2f;
	tempPoly.y[2] = 0.1f;
	tempPoly.x[3] = 0.1f;
	tempPoly.y[3] = 0.1f;

	polygons[3].insert(polygons[3].end(), 1, tempPoly);
	colours[3] = Colours::lightblue;

	for(i=0;i<4;++i)
	{
		tempPoly = *(polygons[i].begin());
		for(j=0;j<4;++j)
		{
			currentPolys[i].x[j] = tempPoly.x[j];
			currentPolys[i].y[j] = tempPoly.y[j];
		}
	}

	setWantsKeyboardFocus(true);
}

//------------------------------------------------------------------------------
DrawingPanel::~DrawingPanel()
{
	/*int i;
	list<Polygon>::iterator it;

	for(i=0;i<4;++i)
	{
		cout << i << ":" << endl;
		for(it=polygons[i].begin();it!=polygons[i].end();++it)
			cout << "\t" << it->time << endl;
	}*/
}

//------------------------------------------------------------------------------
void DrawingPanel::paint(Graphics& g)
{
	int i, j;
	Polygon tempPoly;
	float width = (float)getWidth();
	float height = (float)getHeight();

	//Clear background.
	g.setColour(Colours::white);
	g.fillAll();

	//Draw polygons.
	for(i=0;i<4;++i)
	{
		Path tempPath;

		g.setColour(colours[i]);
		//tempPoly = *(polygons[i].begin());
		tempPoly = currentPolys[i];
		tempPath.startNewSubPath(tempPoly.x[0] * width, tempPoly.y[0] * height);
		tempPath.lineTo(tempPoly.x[1] * width, tempPoly.y[1] * height);
		tempPath.lineTo(tempPoly.x[2] * width, tempPoly.y[2] * height);
		tempPath.lineTo(tempPoly.x[3] * width, tempPoly.y[3] * height);
		tempPath.lineTo(tempPoly.x[0] * width, tempPoly.y[0] * height);
		g.fillPath(tempPath);
	}

	//Draw handles.
	if(showPoints)
	{
		for(i=0;i<4;++i)
		{
			//tempPoly = *(polygons[i].begin());
			tempPoly = currentPolys[i];
			for(j=0;j<4;++j)
			{
				g.setColour(colours[i]);
				g.fillRect((tempPoly.x[j] * width)-4.0f,
						   (tempPoly.y[j] * height)-4.0f,
						   8.0f,
						   8.0f);
				g.setColour(Colours::black);
				g.drawRect((int)((tempPoly.x[j] * width)-4.0f),
						   (int)((tempPoly.y[j] * height)-4.0f),
						   8,
						   8);
			}
		}
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::mouseDown(const MouseEvent& event)
{
	int i, j;

	if(draggedPoly > -1)
		draggedPoly = -1;
	for(i=3;i>=0;--i)
	{
		for(j=0;j<4;++j)
		{
			Rectangle tempRect((int)(currentPolys[i].x[j]*(float)getWidth())-4,
							   (int)(currentPolys[i].y[j]*(float)getHeight())-4,
							   8,
							   8);

			if(tempRect.contains(event.x, event.y))
			{
				draggedPoly = i;
				draggedPoint = j;
				break;
			}
		}

		if(draggedPoly > -1)
			break;
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::mouseUp(const MouseEvent& event)
{
	draggedPoly = -1;
}

//------------------------------------------------------------------------------
void DrawingPanel::mouseDrag(const MouseEvent& event)
{
	if(draggedPoly > -1)
	{
		float tempX, tempY;
		float x = (float)event.x;
		float y = (float)event.y;

		if(x < 0.0f)
			x = 0.0f;
		else if(x > (float)getWidth())
			x = (float)getWidth();

		if(y < 0.0f)
			y = 0.0f;
		else if(y > (float)getHeight())
			y = (float)getHeight();

		tempX = (x/(float)getWidth()) - currentPolys[draggedPoly].x[draggedPoint];
		tempY = (y/(float)getHeight()) - currentPolys[draggedPoly].y[draggedPoint];

		if(event.mods.isCommandDown())
		{
			for(int i=0;i<4;++i)
			{
				currentPolys[draggedPoly].x[i] += tempX;
				currentPolys[draggedPoly].y[i] += tempY;
			}
		}
		else
		{
			currentPolys[draggedPoly].x[draggedPoint] += tempX;
			currentPolys[draggedPoly].y[draggedPoint] += tempY;
		}

		repaint();
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::keyPressed(const KeyPress& key)
{
	if((key.getKeyCode() == KeyPress::F2Key) && key.isCurrentlyDown())
	{
		showPoints = !showPoints;

		repaint();
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::timerCallback()
{
	float tempf;
	int64 currentTicks = Time::getHighResolutionTicks() - startTicks;

	tempf = ((float)currentTicks/(float)ticksPerSecond);
	if(tempf <= length)
	{
		setTime(((float)currentTicks/(float)ticksPerSecond)/length);
		sendChangeMessage(this);
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::setLength(float val)
{
	length = val;
}

//------------------------------------------------------------------------------
void DrawingPanel::setPolygonColour(int index, const Colour& col)
{
	colours[index] = col;
	repaint();
}

//------------------------------------------------------------------------------
void DrawingPanel::setTime(float val)
{
	int i;

	currentPosition = val;

	for(i=0;i<4;++i)
		updateCurrentPolys(i);

	repaint();
}

//------------------------------------------------------------------------------
void DrawingPanel::storePoly(int index)
{
	int i;
	list<Polygon>::iterator it;

	//Remove previous entry at this time point if it exists.
	for(it=polygons[index].begin();it!=polygons[index].end();++it)
	{
		if(it->time == currentPosition)
		{
			polygons[index].erase(it);
			it=polygons[index].begin();
		}
	}

	currentPolys[index].time = currentPosition;
	for(i=0;i<4;++i)
		currentPolys[index].y[i] = 1.0f-currentPolys[index].y[i];
	polygons[index].push_back(currentPolys[index]);
	for(i=0;i<4;++i)
		currentPolys[index].y[i] = 1.0f-currentPolys[index].y[i];

	//Sorting it after's more convenient than worker out where it's supposed to
	//be inserted first.
	polygons[index].sort();
}


//------------------------------------------------------------------------------
void DrawingPanel::startAnimation()
{
	setTime(0.0f);
	ticksPerSecond = Time::getHighResolutionTicksPerSecond();
	startTicks = Time::getHighResolutionTicks();
	startTimer(30);
}

//------------------------------------------------------------------------------
void DrawingPanel::stopAnimation()
{
	stopTimer();
}

//------------------------------------------------------------------------------
void DrawingPanel::clear()
{
	int i, j;
	Polygon tempPoly;

	for(i=0;i<4;++i)
		polygons[i].clear();

	tempPoly.time = 0.0f;

	tempPoly.x[0] = 0.1f;
	tempPoly.y[0] = 0.9f;
	tempPoly.x[1] = 0.2f;
	tempPoly.y[1] = 0.9f;
	tempPoly.x[2] = 0.2f;
	tempPoly.y[2] = 0.8f;
	tempPoly.x[3] = 0.1f;
	tempPoly.y[3] = 0.8f;

	polygons[0].insert(polygons[0].end(), 1, tempPoly);
	colours[0] = Colours::lightblue;

	tempPoly.x[0] = 0.8f;
	tempPoly.y[0] = 0.9f;
	tempPoly.x[1] = 0.9f;
	tempPoly.y[1] = 0.9f;
	tempPoly.x[2] = 0.9f;
	tempPoly.y[2] = 0.8f;
	tempPoly.x[3] = 0.8f;
	tempPoly.y[3] = 0.8f;

	polygons[1].insert(polygons[1].end(), 1, tempPoly);
	colours[1] = Colours::lightblue;

	tempPoly.x[0] = 0.8f;
	tempPoly.y[0] = 0.2f;
	tempPoly.x[1] = 0.9f;
	tempPoly.y[1] = 0.2f;
	tempPoly.x[2] = 0.9f;
	tempPoly.y[2] = 0.1f;
	tempPoly.x[3] = 0.8f;
	tempPoly.y[3] = 0.1f;

	polygons[2].insert(polygons[2].end(), 1, tempPoly);
	colours[2] = Colours::lightblue;

	tempPoly.x[0] = 0.1f;
	tempPoly.y[0] = 0.2f;
	tempPoly.x[1] = 0.2f;
	tempPoly.y[1] = 0.2f;
	tempPoly.x[2] = 0.2f;
	tempPoly.y[2] = 0.1f;
	tempPoly.x[3] = 0.1f;
	tempPoly.y[3] = 0.1f;

	polygons[3].insert(polygons[3].end(), 1, tempPoly);
	colours[3] = Colours::lightblue;

	for(i=0;i<4;++i)
	{
		tempPoly = *(polygons[i].begin());
		for(j=0;j<4;++j)
		{
			currentPolys[i].x[j] = tempPoly.x[j];
			currentPolys[i].y[j] = tempPoly.y[j];
		}
	}
}

//------------------------------------------------------------------------------
void DrawingPanel::setPolygonList(int index,
								  const std::list<Polygon>& polysList)
{
	polygons[index].assign(polysList.begin(), polysList.end());
	currentPolys[index] = *(polygons[index].begin());
}

//------------------------------------------------------------------------------
const list<Polygon>& DrawingPanel::getPolygonList(int index) const
{
	return polygons[index];
}

//------------------------------------------------------------------------------
void DrawingPanel::updateCurrentPolys(int index)
{
	int i;
	float tempf;
	float distance;
	Polygon *tempPoly;
	list<Polygon>::iterator it1;
	list<Polygon>::iterator it2 = polygons[index].end();

	//Get the last stored poly.
	--it2;

	if(it2->time <= currentPosition)
	{
		currentPolys[index] = *it2;
		for(i=0;i<4;++i)
			currentPolys[index].y[i] = 1.0f-currentPolys[index].y[i];
	}
	else
	{
		//Get the 2 polys around the current time.
		it1 = it2;
		while(it1->time > currentPosition)
			--it1;
		it2 = it1;
		++it2;

		//Get the distance in time between the 2.
		distance = it2->time - it1->time;
		tempf = (currentPosition - it1->time)/distance;

		tempPoly = &(currentPolys[index]);
		//Calculate where the currentPoly should be.
		for(i=0;i<4;++i)
		{
			tempPoly->x[i] = it1->x[i] * (1.0f-tempf);
			tempPoly->x[i] += (tempf * it2->x[i]);
			/*tempPoly->y[i] = (1.0f - it1->y[i]) * (1.0f-tempf);
			tempPoly->y[i] += (tempf * (1.0f - it2->y[i]));*/
			tempPoly->y[i] = it1->y[i] * (1.0f-tempf);
			tempPoly->y[i] += (tempf * it2->y[i]);
			tempPoly->y[i] = 1.0f-tempPoly->y[i];
		}
	}
}
